import { introspectionFromSchema } from 'graphql';

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _createForOfIteratorHelperLoose(o, allowArrayLike) {
  var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"];
  if (it) return (it = it.call(o)).next.bind(it);

  if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
    if (it) o = it;
    var i = 0;
    return function () {
      if (i >= o.length) return {
        done: true
      };
      return {
        done: false,
        value: o[i++]
      };
    };
  }

  throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var scalarLocations = function scalarLocations(introspectionQuery, scalars) {
  var wantedScalar = function wantedScalar(n) {
    return scalars.indexOf(n) > -1;
  };

  var inputTypes = introspectionQuery.__schema.types.filter(function (_ref) {
    var kind = _ref.kind;
    return kind === "INPUT_OBJECT";
  }).reduce(function (a, x) {
    a[x.name] = x;
    return a;
  }, {});

  var outputTypes = introspectionQuery.__schema.types.filter(function (_ref2) {
    var kind = _ref2.kind;
    return kind === "OBJECT";
  }).reduce(function (a, x) {
    a[x.name] = x;
    return a;
  }, {});

  var inputFieldPaths = scalars.map(function (type) {
    return {
      type: type,
      fieldType: type,
      path: []
    };
  });
  var outputFieldPaths = [];

  var searchInputScalars = function searchInputScalars(type, current, path) {
    for (var _iterator = _createForOfIteratorHelperLoose(current.inputFields), _step; !(_step = _iterator()).done;) {
      var field = _step.value;
      var fieldType = unpackInputType(field.type);
      if (fieldType === undefined) continue; // ENUMs, others?

      if (fieldType.kind === "SCALAR" && wantedScalar(fieldType.name)) {
        inputFieldPaths.push({
          type: type,
          fieldType: fieldType.name,
          path: path.concat([field.name])
        });
      } else if (fieldType.kind === "INPUT_OBJECT") {
        searchInputScalars(type, inputTypes[fieldType.name], path.concat([field.name]));
      }
    }
  };

  var searchOutputScalars = function searchOutputScalars(type, current, path) {
    if (path.length > 16) return; // dumb loop protection

    for (var _iterator2 = _createForOfIteratorHelperLoose(current.fields), _step2; !(_step2 = _iterator2()).done;) {
      var field = _step2.value;
      var fieldType = unpackOutputType(field.type);
      if (fieldType === undefined) continue; // ENUMs, others?

      if (fieldType.kind === "OBJECT" && fieldType.name.startsWith("__")) continue;

      if (fieldType.kind === "SCALAR" && wantedScalar(fieldType.name)) {
        outputFieldPaths.push({
          type: type,
          fieldType: fieldType.name,
          path: path.concat([field.name])
        });
      } else if (fieldType.kind === "OBJECT") {
        searchOutputScalars(type, outputTypes[fieldType.name], path.concat([field.name]));
      }
    }
  };

  for (var name in inputTypes) {
    searchInputScalars(name, inputTypes[name], []);
  }

  for (var _name in outputTypes) {
    if (_name === "RootMutationType" || _name === "RootQueryType") {
      searchOutputScalars(_name, outputTypes[_name], []);
    }
  }

  var inputScalars = makeTree(inputFieldPaths, true);
  var outputScalars = makeTree(outputFieldPaths, false);
  return {
    inputScalars: inputScalars,
    outputScalars: outputScalars
  };
};

var makeTree = function makeTree(paths, includeTypeName) {
  var scalarTree = {};

  var nest = function nest(map, path, idx, fieldType) {
    if (idx >= path.length - 1) {
      // last element 
      if (path[idx] in map) throw new Error("Path " + path + " already exists in " + map);
      map[path[idx]] = fieldType;
    } else {
      var next = map[path[idx]] || {};
      if (typeof next === 'string') throw new Error("Path " + path + " overlaps with another path in existing map " + map);
      map[path[idx]] = next;
      nest(next, path, idx + 1, fieldType);
    }
  };

  for (var _iterator3 = _createForOfIteratorHelperLoose(paths), _step3; !(_step3 = _iterator3()).done;) {
    var fp = _step3.value;
    var p = includeTypeName ? [fp.type].concat(fp.path) : fp.path;
    nest(scalarTree, p, 0, fp.fieldType);
  }

  return scalarTree;
};

var unpackInputType = function unpackInputType(type) {
  if (type.kind === "SCALAR" || type.kind === "INPUT_OBJECT") return type;
  if (type.kind === "LIST" || type.kind === "NON_NULL") return unpackInputType(type.ofType);
  return;
};

var unpackOutputType = function unpackOutputType(type) {
  if (type.kind === "SCALAR" || type.kind === "OBJECT") return type;
  if (type.kind === "LIST" || type.kind === "NON_NULL") return unpackOutputType(type.ofType);
  return;
};

var plugin = function plugin(schema, _documents, config) {
  if (!config.scalars) throw new Error("graphql-codegen-scalar-locations needs config key 'scalars' with list of scalar names");
  var intro = introspectionFromSchema(schema);
  var locations = scalarLocations(intro, config.scalars);
  return "\nexport type Node = { [key: string]: Node | string };\nexport const scalarLocations : Record<string,Node> = " + JSON.stringify(locations, null, 2) + ";\n";
};

export { plugin };
//# sourceMappingURL=codegen-graphql-scalar-locations.esm.js.map
